#include <unistd.h>
#include <QtCore/QTextStream>
#include <QtGui/QApplication>
#include <QtGui/QMenuBar>
#include <QtGui/QFileDialog>
#include <QtGui/QColor>
#include <QtGui/QPalette>
#include <QtGui/QTextDocument>
#include <QtGui/QMainWindow>
#include <QtGui/QStatusBar>
#include <QtGui/QCheckBox>
#include <QtGui/QVBoxLayout>
#include <QtGui/QListWidget>
#include <QtGui/QListWidgetItem>

#include "ProgramStructureDialog.h"
#include "UnitStructureDialog.h"
#include "PreferencesDialog.h"
#include "FindDialog.h"

#include "editor.h"

namespace loglan {
namespace vlp {

/**
 * Editor constructor. Initializes and sets variables of Loglan Editor.
 *
 * @param argc command line argc parameter
 * @param argv command line argv parameter
 */
Editor::Editor(int argc, char **argv)
{
	setupUi(this);

	if (argc > 1) {
		strcpy(HomeDir, argv[1]);
		fprintf(stderr, "EDIT: HomeDir inited with: %s (%s)\n", HomeDir, argv[1]);
	}

	find_text = "";
	sensitive = FALSE;

	compiler_path.sprintf("%s/%s", HomeDir, "compile/logcomp");
	gen_path.sprintf("%s/%s", HomeDir, "compile/gen");
	file_path.sprintf("%s", HomeDir);
	
	
	connect(editor, SIGNAL(cursorPositionChanged()), this, SLOT(on_editor_cursorPositionChanged()));
}

/**
 * Editor destructor
 */
Editor::~Editor()
{
}

/**
 * Displays line:column position of cursor
 */
void Editor::on_editor_cursorPositionChanged()
{
	QString position;
	int cx;
	int cy;

	cx = editor->textCursor().blockNumber();
	cy = editor->textCursor().columnNumber();
	position.sprintf("%d:%d", cx, cy);

	statusBar()->showMessage(position);
}

/**
 * Loads given file content to the editor.
 *
 * @param fileName Filename of file which will be read.
 */
void Editor::load(const char *fileName)
{
	fname.sprintf("%s", fileName);

	QFile file(fileName);
	if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
		return;

	editor->clear();

	QTextStream textStream(&file);
	editor->append(textStream.readAll());
// 	while (!file.atEnd()) {
// 		editor->append(textStream.readLine());
// 	}

	file.close();

	editor->repaint();
	setWindowTitle(fileName);
}

/**
 * Saves editor content to the given filename.
 * @param fileName File name of the file where content should be saved.
 */
void Editor::save(const char *fileName)
{
	QFile file(fileName);
	if (!file.open(QIODevice::WriteOnly | QIODevice::Text))
		return;

	file.reset();
	file.write(editor->toPlainText().toAscii().data(), editor->toPlainText().length());
	file.close();
}

/**
 * Invokes compiler in the giving mode.
 * @param mode Mode of compilation. Possible values:
 *             - COMP_MODE - compilation mode
 *             - GEN_MODE - program generation mode
 *             - ALL_MODE - compilation & generation mode.
 */
void Editor::compile(int mode)
{
	char cmd[255];
	messages->setReadOnly(FALSE);
	messages->clear();
	messages->repaint();

	/*i = fname.find('.');*/
	/* if (i>=0) {*/

	/*  fn.truncate(i);*/
	switch(mode) {
	case COMP_MODE:
		sprintf(cmd, "%s %s > comp_data!", compiler_path.toStdString().c_str(),
							fname.toStdString().c_str());
		break;
	case GEN_MODE:
		sprintf(cmd, "%s %s > comp_data!", gen_path.toStdString().c_str(),
							fname.toStdString().c_str());
		break;
	case ALL_MODE:
		sprintf(cmd, "%s %s > comp_data!", compiler_path.toStdString().c_str(),
							fname.toStdString().c_str());
		system(cmd);
		sprintf(cmd, "%s %s >> comp_data!", gen_path.toStdString().c_str(),
							fname.toStdString().c_str());
	break;
	}

	system(cmd);
	QFile f("comp_data!");
	if (!f.open(QIODevice::ReadOnly | QIODevice::Text))
		return;

	QTextStream t(&f);
	while (!t.atEnd()) {
		QString s = t.readLine();
		messages->append(s);
	}
	f.close();

	messages->setReadOnly(TRUE);

	messages->repaint();
	unlink("comp_data!");
	/*}*/
}

/**
 * Empties editor content.
 */
void Editor::on_actionNew_triggered()
{
	on_actionClear_all_triggered();
	fname.sprintf("%s", "");
}

/**
 * Displays additional window 
 */
void Editor::on_actionOpen_triggered()
{
	QString fn = QFileDialog::getOpenFileName(this, "Load file", file_path, "*.log");
	if (!fn.isEmpty())
		load(fn.toAscii().data());
}

/**
 * Saves editor content to the file.
 * If content has been read from file, it is written to this file. Otherwise
 * dialog is shown to save content to the specified by user, file.
 */
void Editor::on_actionSave_triggered()
{
	if (fname.isEmpty()) {
		QString fn = QFileDialog::getSaveFileName(this, "Save file", 
							file_path, "*.log");
		if (!fn.isEmpty()) {
			fname.sprintf("%s", fn.toAscii().data());
			save(fn.toAscii().data());
		}
	} else {
		save(fname.toAscii().data());
	}
	setWindowTitle(fname);
}

/**
 * Saves editor content to the file.
 * Forces saving editor content to the new file. Special dialog is shown for
 * that purpose.
 */
void Editor::on_actionSave_as_triggered()
{
	QString fn = QFileDialog::getSaveFileName(this, "Save file as",
							file_path, "*.log");
	if (!fn.isEmpty()) {
		fname.sprintf("%s", fn.toAscii().data());
		save(fn.toAscii().data());
	}
	setWindowTitle(fname);
}
void Editor::on_actionQuit_triggered()
{
	QApplication::instance()->quit();
}
void Editor::on_actionCopy_triggered()
{
	editor->copy();
}
void Editor::on_actionPaste_triggered()
{
	editor->paste();
}
void Editor::on_actionCut_triggered()
{
	editor->cut();
}
void Editor::on_actionClear_all_triggered()
{
	editor->clear();
}

/**
 * Searches for given text in editor content.
 * Displays window to set search parameters. If text is found sets cursor
 * position on it.
 */
void Editor::on_actionFind_triggered()
{
	dialog::FindDialog dialog(this);

	if (dialog.exec()) {
		sensitive = dialog.isCaseSensitive();
		find_text = dialog.getSearchText();
		
		QTextDocument::FindFlags flags = 0;
		
		if (sensitive) {
			flags |= QTextDocument::FindCaseSensitively;
		}

		editor->find(find_text, flags);
	}
}

/**
 * Searches for next occurence of given text in editor content.
 * Displays window to set search parameters. If text is found sets cursor
 * position on it.
 */
void Editor::on_actionFind_Next_triggered()
{
	if (!find_text.isEmpty()) {
		QTextDocument::FindFlags flags = 0;
		
		if (sensitive) {
			flags |= QTextDocument::FindCaseSensitively;
		}
		editor->find(find_text, flags);
	}
}

/**
 * Displays window with editor properties
 */
void Editor::on_actionPreferences_triggered()
{
	dialog::PreferencesDialog dialog(this);

	dialog.setCompilerPath(compiler_path);
	dialog.setFilesPath(file_path);
	dialog.setGenPath(gen_path);

	if (dialog.exec()) {
		compiler_path.sprintf("%s", dialog.getCompilerPath().toStdString().c_str());
		gen_path.sprintf("%s",dialog.getGenPath().toStdString().c_str());
		file_path.sprintf("%s", dialog.getFilesPath().toStdString().c_str());
	};
}

/**
 * Saves and compiles code.
 */
void Editor::on_actionCompile_triggered()
{
	on_actionSave_triggered();
	compile(COMP_MODE);
}

/**
 * Generates program.
 */
void Editor::on_actionGen_triggered()
{
	compile(GEN_MODE);
}

/**
 * Saves, compiles and generates code.
 */
void Editor::on_actionCompile_Gen_triggered()
{
	on_actionSave_triggered();
	compile(ALL_MODE);
}

void Editor::on_actionProgram_structure_triggered()
{
	dialog::ProgramStructureDialog dialog(this);

	if (dialog.exec()) {
		editor->textCursor().insertText(dialog.getCode());
	}
}

void Editor::on_actionUnit_structure_triggered()
{
	dialog::UnitStructureDialog dialog(this);

	if (dialog.exec()) {
		editor->textCursor().insertText(dialog.getCode());
	}
}

}
}

/**
 * Program main function.
 * argv[1] is mandatory and should be a path to the home directory of
 * application (same as in configuration file).
 *
 * @param argc Number of program arguments
 * @param argv Program arguments
 */
int main(int argc, char **argv)
{
	QApplication app(argc, argv);
	loglan::vlp::Editor * editor = new loglan::vlp::Editor(argc, argv);
	editor->show();
	return app.exec();
}

