#include "vlp/config.h"

#include <cstring>

namespace loglan {
namespace vlp {

Config::Config()
{
	config_init(&cfg);
	file = NULL;
}

Config::~Config()
{
	config_destroy(&cfg);
	if (file) {
		fclose(file);
		file = NULL;
	}
}

bool Config::load(const char * configFilePath)
{
	file = fopen(configFilePath, "rt");
	if (!file) {
		fprintf(stderr,
			"Error: Cannot load configuration file: %s!\n",
			configFilePath
		);
		return false;
	}
	
	if (!config_read(&cfg, file)) {
		fprintf(stderr,
			"%s! In file %s, line %d\n",
			config_error_text(&cfg),
			config_error_file(&cfg),
			config_error_line(&cfg));
		return false;
	}

	return true;
}

bool Config::save(const char * configFilePath)
{
	if(!config_write_file(&cfg, configFilePath)) {
		fprintf(stderr,
			"Error: Cannot save configuration file: %s!\n",
			configFilePath
		);
		return false;
	}

	return true;
}

int Config::getNodeNumber()
{
	config_setting_t *setting = config_lookup(&cfg, "node_number");

	if (!setting) {
		fprintf(stderr,
			"%s! In file %s, '%s' was not found.\n",
			"Warning",
			config_error_file(&cfg),
			"node_number"
		);

		throw;
	}

	return config_setting_get_int(setting);
}
void Config::setNodeNumber(int nodeNumber)
{
	config_setting_t *setting;
	config_setting_t *root = config_root_setting(&cfg);

	setting = config_setting_get_member(root, "node_number");

	if(!setting) {
		setting = config_setting_add(
			root,
			"node_number",
			CONFIG_TYPE_INT
		);
	}
	config_setting_set_int(setting, nodeNumber);
}

std::vector<std::string> Config::getHosts()
{
	config_setting_t *setting = config_lookup(&cfg, "host");
	std::vector<std::string> hosts;

	if (!setting) {
		fprintf(stderr,
			"%s! In file %s, '%s' was not found.\n", 
			"Warning",
			"configuration",//config_error_file(&cfg),
			"host"
		);

		return hosts;
	}

	switch(config_setting_type(setting)) {
	/* TODO: Deprecated. Made for back compatibility. */
	case CONFIG_TYPE_STRING: {
		std::string host(config_setting_get_string(setting));
		hosts.push_back(host);
		break;
	}
	case CONFIG_TYPE_ARRAY: {
		int size = config_setting_length(setting);
		std::string host;
		for (int i = 0; i < size; i++) {
			host = config_setting_get_string_elem(setting, i);
			hosts.push_back(host);
		}
		break;
	}
	default:
		fprintf(stderr,
			"%s! In file %s, bad entry type for %s will not be read.\n",
			"Warning",
			"configuration",//config_error_file(&cfg),
			"host"
		);
	}

	return hosts;
}

void Config::setHosts(std::vector<std::string> hosts)
{
	config_setting_t *root = config_root_setting(&cfg);
	config_setting_t *setting = NULL;

	setting = config_setting_add(root, "host", CONFIG_TYPE_ARRAY);

	for (auto host : hosts) {
		config_setting_set_string(
			config_setting_add(setting, NULL, CONFIG_TYPE_STRING),
			host.c_str()
		);
	}
}

ConnectionType Config::getConnectionType()
{
	config_setting_t *setting = config_lookup(&cfg, "type");

	if (!setting) {
		fprintf(stderr,
			"%s! In file %s, '%s' was not found.\n",
			"Warning",
			"configuration",//config_error_file(&cfg),
			"type"
		);

		throw;
	}

	/* same as strcmp(..) == 0 */
	if (!strcmp(config_setting_get_string(setting), "explicit")) {
		return EXPLICIT;
	}

	return REGISTER;
}
void Config::setConnectionType(ConnectionType type)
{
	config_setting_t *setting;
	config_setting_t *root = config_root_setting(&cfg);

	setting = config_setting_get_member(root, "type");

	if(!setting) {
		setting = config_setting_add(
			root,
			"type",
			CONFIG_TYPE_STRING
		);
	}

	if (type == EXPLICIT) {
		config_setting_set_string(setting, "explicit");
	}
	else if (type == REGISTER) {
		config_setting_set_string(setting, "register");
	}
}

const char * Config::getProgramDir()
{
	config_setting_t *setting = config_lookup(&cfg, "progdir");

	if (!setting) {
		fprintf(stderr,
			"%s! In file %s, '%s' was not found.\n",
			"Warning",
			"configuration",//config_error_file(&cfg),
			"progdir"
		);
	}

	return config_setting_get_string(setting);
}
void Config::setProgramDir(const char * programDir)
{
	config_setting_t *setting;
	config_setting_t *root = config_root_setting(&cfg);

	setting = config_setting_get_member(root, "progdir");

	if(!setting) {
		setting = config_setting_add(
			root,
			"progdir",
			CONFIG_TYPE_STRING
		);
	}
	config_setting_set_string(setting, programDir);
}

}
}