/**************************************************************

  Copyright (C) 1997  Oskar Swida

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.




 NOTE: This software is using the free software license of 
       the QT library v.1.30 from Troll Tech AS.
       See the file LICENSE.QT.

 
  To contact the author, write:
     e-mail: swida@aragorn.pb.bialystok.pl

************************************************************/

#include <QtGui/QApplication>
#include <QtGui/QMainWindow>
#include <QtGui/QTextEdit>
#include <QtGui/QMenuBar>
#include <QtGui/QMessageBox>
#include <QtGui/QFileDialog>
#include <QtGui/QDialog>
#include <QtCore/QString>
#include <QtGui/QLabel>
#include <QtGui/QLineEdit>
#include <QtGui/QPushButton>
#include <QtGui/QRadioButton>
#include <QtGui/QGroupBox>
#include <QtGui/QVBoxLayout>
#include <QtCore/QList>
#include <QtGui/QListWidget>
#include <qfile.h>
#include <qcursor.h>
#include <QtCore/QSocketNotifier>
#include <QtGui/QCloseEvent>
#include <QtCore/QDir>
#include <QtCore/QProcess>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>

#include "genint1.h"
#include "socu.h"
#include <netinet/in.h>

#include <libconfig.h>
#include <X11/Xlib.h>

#include "kernel.h"
#include "ui/KernelWindow.h"
#include "ConnectDialog.h"
#include "KillDialog.h"
#include "LockDialog.h"
#include "OptionsDialog.h"
#include "MessageDialog.h"

#include "vlp/config.h"
#include "vlp/QtConfigurationFinder.h"
#include "vlp/exception/ConfigFileNotFound.h"

#include <sys/prctl.h>
/* File resides in top directory (where are Makefiles)*/
#include "../../config.h"


namespace loglan {
namespace vlp {

char CharLine[25] = "________________________";

/**
 * Event invoked on program close.
 * Closes application. Displays additional window to confirm exit.
 */
void QKernel::closeEvent(QCloseEvent * e)
{
	e->ignore();

	if (!LOCKED) {
		on_actionQuit_triggered();
	}
}

void QKernel::setLocked(bool locked)
{
	LOCKED = locked;

	actionQuit->setDisabled(locked);

	actionExecute->setDisabled(locked);
	actionKill->setDisabled(locked);
	actionMessage->setDisabled(locked);
	actionConnect->setDisabled(locked);
	actionDisconnect->setDisabled(locked);
	actionInfo->setDisabled(locked);


	/* Enable only menu entry for unlocking */
	actionEditor->setDisabled(locked);
	actionOptions->setDisabled(locked);
	actionLock_console->setDisabled(locked);
	actionUnlock_console->setDisabled(!locked);
}

/**
 * Kernel program constructor.
 * Prepares everything to work.
 */
QKernel::QKernel(int argc, char **argv)
{
	setupUi(this);

	int i;
	for(i = 0; (i < 5) && (i < argc-1); i++) {
		strcpy(myargs[i], "");
		if (i < argc) {
			strcpy(myargs[i], argv[i+1]);
		}
	}

	loglan::vlp::QtConfigurationFinder configFinder;
	configFinder.initSearchDirs();

	QDir q(getRemoteDir());

	if (!q.exists()) {
		q.mkpath(getRemoteDir());
	}

	info_messages = actionInfo_messages->isChecked();
	wait_for_info = FALSE;

	setWindowTitle(PACKAGE_NAME);

	LOCKED = FALSE;

	Tasks = 0;
	freeINTid = 1;
	ActiveConnections = 0;
	strcpy(LockPasswd, "");

	loadConfig(configFinder.findConfig().c_str());

	RunNetModule();

	Net_Notify = new QSocketNotifier(net_sock, QSocketNotifier::Read, this);
	connect(Net_Notify, SIGNAL(activated(int)), this, SLOT(NetMessage()));

	WriteMessage("\n " PACKAGE_STRING ": READY \n");
}

QString QKernel::getConfigFilePath()
{
	return homeDir.absoluteFilePath("vlp.cfg");
}

const char * QKernel::getHomeDir()
{
	return homeDir.absolutePath().toStdString().c_str();
}

const char * QKernel::getRemoteDir()
{
	return homeDir.absoluteFilePath(REMOTE_PATH).toStdString().c_str();
}

const char * QKernel::getNetModuleSocket()
{
	return homeDir.absoluteFilePath(NPATH).toStdString().c_str();
}

const char * QKernel::getGraphModuleSocket()
{
	return homeDir.absoluteFilePath(GPATH).toStdString().c_str();
}

void QKernel::loadConfig(const QString & fname)
{
	loadConfig(fname.toStdString().c_str());
}
/**
 * Loads configuration from the given file.
 * @param fname Filename of the configuration file.
 */
void QKernel::loadConfig(const char * fname)
{
	loglan::vlp::Config config;
	if(config.load(fname)) {
		NodeNumber = config.getNodeNumber();
		ConType = config.getConnectionType();


		std::vector<std::string> hosts = config.getHosts();
		for (unsigned int i = 0; i < hosts.size(); i++) {
			ConnectList.append(new ConnectEntry(hosts[i].c_str()));
		}

		strncpy(progdir, config.getProgramDir(), 256);

		homeDir = QDir(QCoreApplication::applicationDirPath());
	}
}

/**
 * Executes program.
 * Additional window id displayed to set which code to execute.
 */
void QKernel::on_actionExecute_triggered()
{
	QString s = QFileDialog::getOpenFileName(this, "Execute", progdir, "*.log");

	if (!s.isNull()) {
		int i = s.indexOf(".log");

		if (i > 0)
			s.remove(i, 4);

		// @TODO: if no interpreter is running will result in killing app
		RunIntModule((char*)s.toAscii().data(), 0);
	}
}

/**
 * Invokes editor program
 */
void QKernel::on_actionEditor_triggered()
{
	QString program = getHomeDir();
	program += "/modules/logedit";

	fprintf(stderr, "Run EDIT Module: %s, %s\n", program.toStdString().c_str(), getHomeDir());
	pid_t pid = fork();
	if (pid == 0) {
		if (execl(program.toStdString().c_str(),
			"logedit",
			getHomeDir(),
			myargs[0],
			myargs[1],
			myargs[2],
			myargs[3],
			myargs[4],
			NULL
		) == -1) {
			WriteMessage("Executing logedit failed!");
		}
	}
	else if (pid < 0) {
		WriteMessage("fork(logedit) failed!");
		WriteMessage("Exiting...");
		sleep(2);
		on_actionQuit_triggered();
		exit(3);
	}
}

/**
 * Invokes help program
 */
void QKernel::on_actionHelp_triggered()
{
	QString program = getHomeDir();
	program += "/modules/loghelp";
	
	QString docDir = getHomeDir();
	docDir += "/doc";

	fprintf(stderr, "Run HELP Module: %s, %s\n", program.toStdString().c_str(), docDir.toStdString().c_str());

	pid_t pid = fork();
	if (pid == 0) {
		if (execl(program.toStdString().c_str(),
			"loghelp",
			docDir.toStdString().c_str(),
			myargs[0],
			myargs[1],
			myargs[2],
			myargs[3],
			myargs[4],
			NULL
		) == -1) {

			WriteMessage("Executing loghelp failed!");
		}
	}
	else if (pid < 0) {
		WriteMessage("fork(loghelp) failed!");
		WriteMessage("Exiting...");
		sleep(2);
		on_actionQuit_triggered();
		exit(3);
	}
}

/**
 * Invokes graphics module
 */
void QKernel::RunGraphModule(char *sk)
{
	QString program = getHomeDir();
	program += "/modules/loggraph";

	fprintf(stderr, "Run GRAPH Module: %s, %s\n", program.toStdString().c_str(), sk);

	pid_t pid = fork();
	if (pid == 0) {
		if (execl(program.toStdString().c_str(),
			"loggraph",
			program.toStdString().c_str(),
			sk,
			myargs[0],
			myargs[1],
			myargs[2],
			myargs[3],
			myargs[4],
			NULL
		) == -1) {

			WriteMessage("Executing loggraph failed!");
			WriteMessage("Exiting...");
			sleep(2);
			on_actionQuit_triggered();
		}
	}
	else if (pid < 0) {
		WriteMessage("fork(loggraph) failed!");
		WriteMessage("Exiting...");
		sleep(2);
		on_actionQuit_triggered();
		exit(3);
	}
}

/**
 * Invokes net module
 */
void QKernel::RunNetModule()
{

	struct sockaddr_un svr;
	int len;
	int on;
	int sock;

	QString program = getHomeDir();
	program += "/modules/lognet";

	fprintf(stderr, "Run NET Module: %s, %s %s\n", program.toStdString().c_str(), getNetModuleSocket(), getConfigFilePath().toStdString().c_str());
	pid_t pid = fork();
	if (pid == 0) {
		if (execl(program.toStdString().c_str(),
			"lognet",
			getNetModuleSocket(),
			getConfigFilePath().toStdString().c_str(),
			myargs[0],
			myargs[1],
			myargs[2],
			myargs[3],
			myargs[4],
			NULL
		) == -1) {

			WriteMessage("Executing lognet failed!");
			WriteMessage("Exiting...");
			sleep(2);
			on_actionQuit_triggered();
		}
	}
	else if (pid < 0) {
		WriteMessage("fork(lognet) failed!");
		WriteMessage("Exiting...");
		sleep(2);
		on_actionQuit_triggered();
		exit(3);
	}

	/* -------- socket for NET module -------- */
	unlink(getNetModuleSocket());
	sock = socket(AF_UNIX, SOCK_STREAM, 0);
	bzero(&svr, sizeof(svr));
	svr.sun_family = AF_UNIX;
	strcpy(svr.sun_path, getNetModuleSocket());
	len = strlen(svr.sun_path) + sizeof(svr.sun_family);
	bind(sock, (struct sockaddr*)&svr, len);
	listen(sock, 5);

	net_sock = accept(sock, (struct sockaddr*)0, (unsigned int*)0);
	// close(sock); 
	if (net_sock != 0) {
		WriteMessage("NETWORK successfully connected");
		fcntl(net_sock, F_SETFL, O_NONBLOCK|fcntl(net_sock, F_GETFL, 0));
		on = 1;
		setsockopt(net_sock, IPPROTO_TCP, TCP_NODELAY, (char*)&on, sizeof(on));
	} else {
		WriteMessage("Cannot connect NETWORK resources");
		WriteMessage("Exiting...");
		sleep(2);
		on_actionQuit_triggered();
	}
}

/**
 * Connects to the specified address
 * Additional window is displayed to connect to the specified address
 */
void QKernel::on_actionConnect_triggered()
{
	MESSAGE m;

	dialog::ConnectDialog dialog(this);

	if (dialog.exec()) {
		m.msg_type = MSG_NET;
		m.param.pword[0] = NET_CONNECT_TO;
		strcpy(m.param.pstr, dialog.getAddress().toStdString().c_str());
		write(net_sock, &m, sizeof(MESSAGE));
	}
}

/**
 * Disconnects from virtual machine
 */
void QKernel::on_actionDisconnect_triggered()
{
	MESSAGE msg;

	if (info_messages)
		WriteMessage("Disconnecting from virtual machine");

	msg.msg_type = MSG_NET;
	msg.param.pword[0] = NET_DISCONNECT;
	write(net_sock, &msg, sizeof(MESSAGE));
}

/**
 * Quits process. Closes VLP. Shows additional window to confirm exit. 
 */
void QKernel::on_actionQuit_triggered()
{
	MESSAGE msg;

	QMessageBox::StandardButton response = QMessageBox::question(this,
		"Close VLP",
		"Terminate VLP ?",
		QMessageBox::Ok | QMessageBox::Cancel
	);


	if (response == QMessageBox::Cancel) {
		return;
	}
	/*
	msg.msg_type = MSG_NET;
	msg.param.pword[0] = NET_DISCONNECT;
	write(net_sock, &msg, sizeof(MESSAGE));*/
	delete Net_Notify;

	msg.msg_type = MSG_NET;
	msg.param.pword[0] = NET_EXIT;
	write(net_sock, &msg, sizeof(MESSAGE));
	/*  ::close(net_sock);*/
	QApplication::instance()->quit();
}

/**
 * Sends message to node.
 * Additional window is displayed to set Node Number of node where send message,
 * and textfield to enter message.
 */
void QKernel::on_actionMessage_triggered()
{
	dialog::MessageDialog dialog(this);
	MESSAGE m;

	if (dialog.exec()) {
		m.msg_type = MSG_NET;
		m.param.pword[0] = NET_PROPAGATE;
		m.param.pword[1] = MSG_VLP;
		m.param.pword[2] = NodeNumber;
		m.param.pword[4] = dialog.getNodeNumber();
		m.param.pword[6] = VLP_WRITE;
		strcpy(m.param.pstr, dialog.getMessage().toStdString().c_str());
		write(net_sock, &m, sizeof(MESSAGE));
	}
}

/**
 * Kills interpreter.
 * Additional window is displayed to get ID of interpreter which should be
 * killed.
 */
void QKernel::on_actionKill_triggered()
{
	dialog::KillInterpreterDialog dialog(this);

	if (dialog.exec()) {
		MESSAGE m;
		m.msg_type = MSG_INT;
		m.param.pword[0] = INT_KILL;
		
		InterpEntry *interpreter;
		interpreter = findINTbyID(dialog.getInterpreterId());

		if (interpreter != NULL) {
			if (!(interpreter->remote)) {
				write(interpreter->sock, &m, sizeof(MESSAGE));
			}
			else {
				WriteMessage("This is a remote instance of a program!");
			}
		}
		else {
			WriteMessage("Interpreter not found");
		}
	}
}

/**
 * Sends message to the net module.
 *
 * \todo method needs to be refactored
 */
void QKernel::NetMessage()
{
	
	/* TODO: It has to be rewritten */
	MESSAGE msg;
	int cnt;
	InterpEntry *pom;

	cnt = read(net_sock, &msg, sizeof(MESSAGE));
	if ((cnt > 0) && (msg.msg_type == MSG_NET)) {
		char ss[255];

		switch(msg.param.pword[0]) {
		case NET_CSWRITELN:
			WriteMessage(msg.param.pstr);
			break;
		case NET_PROPAGATE: 
			switch(msg.param.pword[1]) {
			case MSG_INT:
				/*  pom = find_link_by_ID(msg.param.pword[5]);
				msg.msg_type = MSG_NET;
				msg.param.pword[0] = NET_PROPAGATE;
				send_int(pom, &msg);*/
				break;
			case MSG_VLP:
				switch(msg.param.pword[6]) {
				case VLP_WRITE:
					QApplication::beep();
					WriteMessage(CharLine);
					WriteMessage("### Incoming Messsage ###");
					sprintf(ss,
						"Mesg from Node %d: %s",
						msg.param.pword[2],
						msg.param.pstr
					);
					WriteMessage(ss);
					WriteMessage(CharLine);
					break;
				case VLP_REMOTE_INSTANCE:
					sprintf(ss, "%s/%s", getRemoteDir(), msg.param.pstr);

					if (info_messages) { 
						WriteMessage("Running program:");
						WriteMessage(ss);
					}
					pom = RunIntModule(ss, 1);
					if (pom != NULL) {
						pom->p_ctx.node = msg.param.pword[2];
						pom->p_ctx.program_id = 
							msg.param.pword[7];
						pom->RInstances[msg.param.pword[2]] = msg.param.pword[7];
					}
					break;
				case VLP_CLOSE_INSTANCE:
					msg.msg_type = MSG_INT;
					msg.param.pword[0] = INT_CLOSE_INSTANCE;
					pom = findINTbyID(msg.param.pword[7]);
					if (pom != NULL) {
						write(pom->sock, &msg,
							sizeof(MESSAGE));
						MESSAGE m1;
						m1.msg_type = MSG_VLP;
						m1.param.pword[0] = VLP_INTERPRETER_DOWN;
						m1.param.pword[1] = pom->ID;
						write(net_sock, &m1,
							sizeof(MESSAGE));
					} else {
						WriteMessage("Instance not found"); 
					}
					break; 
				} /* VLP switch */
			}/* switch */
			break;
		case NET_CONNECTIONS:
			ActiveConnections = msg.param.pword[1];
			WriteMessage(msg.param.pstr);
			if (!synchro) 
				synchro = TRUE;
			break;
		case NET_INFO:
			/* TODO: It has to be rewritten */
			if (wait_for_info) {
				QString poms, poms1, poms2;
				poms.sprintf("%s", msg.param.pstr);
				while (poms.length() > 0) {
					cnt = poms.indexOf(';');
					if (cnt!=-1) {
						poms1 = poms.left(cnt);
						poms = poms.right(poms.length() - cnt - 1);
						cnt = poms1.indexOf('=');
						if (cnt != -1) {
							poms2 = poms1.left(cnt);
							poms1 = poms1.right(poms1.length() - cnt - 1);
							sprintf(ss, "Node: %s Addr: %s", poms2.toStdString().c_str(), poms1.toStdString().c_str());
							WriteMessage(ss); 
						}
					} 
				}
			} 
			break;
		case NET_INFO_END:
			wait_for_info = FALSE;
			WriteMessage(CharLine);
			break;
		} /* switch */
	}
}

/**
 * Sends message to the interpreter program.
 * @param sock Interpreter socket to whom the message will be send.
 */
void QKernel::IntMessage(int sock)
{
	MESSAGE msg;
	int cnt;
	InterpEntry *e;

	cnt = read(sock, &msg, sizeof(MESSAGE));
	e = findINTbySocket(sock);
	if ((cnt > 0) && (e != NULL)) {
		switch (msg.msg_type) { 
		case MSG_GRAPH:
			if (msg.param.pword[0] == GRAPH_ALLOCATE) {
				fprintf(stderr, "Running graph module with arg: %s\n", msg.param.pstr);
				RunGraphModule(msg.param.pstr);
			}
			break;
		case MSG_NET:
			write(net_sock, &msg, sizeof(MESSAGE));
			break;
		case MSG_VLP:
			switch(msg.param.pword[0]) {
			case VLP_REMOTE_INSTANCE_PLEASE:
				RemoteInstance(e, msg.param.pword[2]);
				break;
			}/* switch */
			break;
		case MSG_INT:
			switch(msg.param.pword[0]) {
			case INT_EXITING:
				MESSAGE m;
				m.msg_type = MSG_VLP;
				m.param.pword[0] = VLP_INTERPRETER_DOWN;
				m.param.pword[1] = e->ID;
				write(net_sock, &m, sizeof(MESSAGE));
				if (e->remote == 0)
					CloseInstances(e);
				delete e->notify;
				::close(e->sock);
				/* TODO: Check this */
				Interpreters.removeOne(e);
				delete e;

				if (info_messages) {
					char ss[255];
					sprintf(ss, "%s : End of program "
						"execution", msg.param.pstr);
					WriteMessage(ss);
				}
				break;
			case INT_CTX_REQ:
				msg.msg_type = MSG_INT;
				msg.param.pword[0] = INT_CTX;
				msg.param.pword[1] = NodeNumber;
				msg.param.pword[2] = e->ID;
				if (e->remote) {
					msg.param.pword[3] = e->p_ctx.node;
					msg.param.pword[4] = 
							e->p_ctx.program_id;
				}
				write(sock, &msg, sizeof(MESSAGE)); 
				break;
			};
			break;
		}
	}
}

/**
 * Writes message to kernel logger.
 * @parame msg String with message to log
 */
void QKernel::WriteMessage(const char *msg)
{
	int x;
// 	int y;
	x = desktop->textCursor().blockNumber();
// 	y = desktop->textCursor().columnNumber();

	if (x > 100) {
		desktop->clear();
	}
	
	desktop->setReadOnly(FALSE);
	desktop->append(msg);
	desktop->setReadOnly(TRUE);

	QTextCursor tmpCursor = desktop->textCursor();
	tmpCursor.movePosition(QTextCursor::End, QTextCursor::MoveAnchor);
	desktop->setTextCursor(tmpCursor);

	desktop->repaint();
	
	if (desktop->document()->blockCount() > 100) {
		desktop->clear();
	}
}

/**
 * Adds checkbox to menu item. If it is checked additional info messages are
 * shown.
 */
void QKernel::on_actionInfo_messages_triggered()
{
	if (toolsMenu != NULL) {
		info_messages = actionInfo_messages->isChecked();
		actionInfo_messages->setChecked(info_messages);
	}
}

/**
 * Allows to set options in GUI window.
 * Additional window is displayed to set kernel options which are saved in 
 * vlp.cfg file in kernel executable directory.
 */
void QKernel::on_actionOptions_triggered()
{
	dialog::OptionsDialog optionsDialog(getConfigFilePath(), this);
	if (optionsDialog.exec()) {
		optionsDialog.saveConfig(getConfigFilePath());

		loadConfig(getConfigFilePath());
	}
}

/**
 * Locks kernel program.
 * Additional window is displayed to enter password and retype it. If both are
 * same kernel window is locked.
 */
void QKernel::on_actionLock_console_triggered()
{
	dialog::LockDialog lockDialog(this);

	if (lockDialog.exec()) {
		QString password = lockDialog.getPassword();
		if (lockDialog.getPassword().size() > 0) {
			strcpy(LockPasswd, password.toStdString().c_str());
			lockDialog.retype();

			if (lockDialog.exec()) {
				password = lockDialog.getPassword();
				if (password == LockPasswd) {
					setLocked(TRUE);
					WriteMessage("CONSOLE LOCKED");
				} else {
					QMessageBox msg(this);
					msg.setText("Not matching!");
					msg.setButtonText(0, "Close");
					msg.show();
					strcpy(LockPasswd, "");
				}
			} else {
				strcpy(LockPasswd, "");
			}
		}
	}
}

/**
 * Unlocks kernel program.
 * Additional window is displayed to enter password. If it is correct, kernel 
 * window is unlocked
 */
void QKernel::on_actionUnlock_console_triggered()
{
	dialog::LockDialog lockDialog(this);

	if (lockDialog.exec()) {
		QString password = lockDialog.getPassword();
		if (strcmp(password.toStdString().c_str(), LockPasswd) == 0) {
			setLocked(FALSE);
			WriteMessage("CONSOLE UNLOCKED");
		} else {
			QMessageBox msg(this);
			msg.setText("Wrong password!");
			msg.setButtonText(0, "Close");
			msg.show();
		}
	}
}

/**
 * Finds Interpreter by its socket
 * @param _id ID of the socket
 * @return returns pointer to the found interpreter slot. NULL otherwise
 */
InterpEntry *QKernel::findINTbySocket(int _id)
{
	InterpEntry *pom = NULL;
	
	for (auto interpreter : Interpreters) {
		if (interpreter->sock == _id) {
			pom = interpreter;
			break;
		}
	}

	return pom;
}

/**
 * Finds Interpreter by its ID.
 * @param _id ID of the interpreter
 * @return returns pointer to the found interpreter slot. NULL otherwise
 */
InterpEntry *QKernel::findINTbyID(int _id)
{
	InterpEntry *pom = NULL;
	
	for (auto interpreter : Interpreters) {
		if (interpreter->ID == _id) {
			pom = interpreter;
			break;
		}
	}

	return pom;
}


/**
 * Connects interpreter
 * @param ss full filepath with filename but without extension of the loglan
 *           program to run.
 * @param r Interpreter execution mode. 0 if it will be local instance, 1 if
 *          remote
 * @return Returns pointer to newly created interpreter slot, or NULL on error.
 */
InterpEntry *QKernel::RunIntModule(char *ss, int r)
{
	fprintf(stderr, "Run INT Module: %s, %d\n", ss, r);
	char a[256], b[255];
	struct sockaddr_un svr;
	int len, sock, on;
	unsigned int i;
	int newint=-1;
	char cmd[255];
	FILE *cf;
	MESSAGE msg;
	InterpEntry *newINT = NULL;

	strcpy(a, ss);
	strcat(a, ".ccd");
	cf = fopen(a, "r");
	if (cf == NULL) {
		WriteMessage("File not found: no .ccd file");
		return NULL;
	}
	fclose(cf);

	strcpy(a, ss);
	strcat(a, ".pcd");
	cf = fopen(a, "r");
	if (cf == NULL) {
		WriteMessage("File not found: no .pcd file");
		return NULL;
	}
	fclose(cf);

	newINT = new InterpEntry;
	for(i = 0; i < MAXINSTANCES; i++) 
		newINT->RInstances[i] =- 1;

	strcpy(b, rindex(ss, '/'));
	for(i = 0; i < strlen(b); i++)
		b[i] = b[i+1];
	if (info_messages) {
		sprintf(a, "%s : Start execution", b);
		WriteMessage(a); 
	}

	newint = freeINTid;
	freeINTid++;
	newINT->ID = newint;
	strcpy(newINT->shortname, b);
	strcpy(newINT->fullname, ss);

	sprintf(a, "%s%d", homeDir.absoluteFilePath(IPATH).toStdString().c_str(), newint);
	sprintf(cmd, "%s/modules/logint %s %s",
		getHomeDir(),
		a,
		ss);
	if (r) {
		strcat(cmd, " r");
	}
// 	sprintf(b, " %s %s %s %s %s",
// 		myargs[0], myargs[1], myargs[2], myargs[3], myargs[4]);
// 	sprintf(cmd, "%s %s", cmd, b);
// 	strcat(cmd, b);
	fprintf(stderr, "%s\n", cmd);
	strcat(cmd, " &");

	sock = socket(AF_UNIX, SOCK_STREAM, 0);
	unlink(a);
	bzero(&svr, sizeof(svr));
	svr.sun_family = AF_UNIX;
	strcpy(svr.sun_path, a);
	len = strlen(svr.sun_path)+sizeof(svr.sun_family);
	bind(sock, (struct sockaddr*)&svr, len);
	listen(sock, 5);
	system(cmd);
	newINT->sock = accept(sock, (struct sockaddr*)0, (unsigned int *)0);
	//::close(sock);

	if (newINT->sock > 0) { 
		fcntl(newINT->sock, F_SETFL, 
				O_NONBLOCK|fcntl(newINT->sock, F_GETFL, 0));
		on=1; 
		setsockopt(newINT->sock, IPPROTO_TCP, TCP_NODELAY, (char*)&on,
								sizeof(on));
		if (r)
			newINT->remote = 1;
		else 
			newINT->remote = 0;

		bzero(&msg, sizeof(MESSAGE));
		msg.msg_type = MSG_VLP;
		msg.param.pword[0] = VLP_REGINT;
		msg.param.pword[1] = newINT->ID;
		sprintf(msg.param.pstr, "logi%d.net", newint);
		write(net_sock, &msg, sizeof(MESSAGE)); 

		Interpreters.append(newINT);
		newINT->notify = new QSocketNotifier(newINT->sock,
							QSocketNotifier::Read);
		connect(newINT->notify, SIGNAL(activated(int)), this,
							SLOT(IntMessage(int)));
		if (info_messages)
			WriteMessage("INTERPRETER successfully connected");
	} else {
		WriteMessage("Cannot connect interpreter");
	}

	return newINT;
}

/**
 * Allocates remote instance of interpreter
 * @param interp Interpreter slot 
 * @param on Node Number
 */
void QKernel::RemoteInstance(InterpEntry *interp, int on)
{
	MESSAGE m;

	m.msg_type = MSG_NET;
	m.param.pword[0] = NET_NODE_EXIST;
	m.param.pword[1] = on;
	m.param.pword[2] = interp->ID;
	write(net_sock, &m, sizeof(MESSAGE));
	bzero(&m, sizeof(MESSAGE));
	while((m.msg_type!=MSG_NET) && (m.param.pword[0]!=NET_NODE_EXIST))
		read(net_sock, &m, sizeof(MESSAGE));

	 /* means node exists */
	if (m.param.pword[1] == 1) {
		m.msg_type = MSG_NET;
		m.param.pword[0] = NET_TRANSMIT_CODE;
		m.param.pword[1] = interp->ID;
		m.param.pword[2] = on;
		strcpy(m.param.pstr, interp->fullname);
		write(net_sock, &m, sizeof(MESSAGE));

		Net_Notify->setEnabled(FALSE);
		while ((m.msg_type != MSG_NET) ||
			(m.param.pword[0] != NET_TRANSMITTED))
			read(net_sock, &m, sizeof(MESSAGE));

		m.msg_type = MSG_NET;
		m.param.pword[0] = NET_PROPAGATE;
		m.param.pword[1] = MSG_VLP;
		m.param.pword[2] = NodeNumber;
		m.param.pword[3] = 0;
		m.param.pword[4] = on;
		m.param.pword[5] = 0;
		m.param.pword[6] = VLP_REMOTE_INSTANCE;
		m.param.pword[7] = interp->ID;
		strcpy(m.param.pstr, interp->shortname);
		write(net_sock, &m, sizeof(MESSAGE));

		read(net_sock, &m, sizeof(MESSAGE));
		while (1) {
			if ((m.param.pword[0] == NET_PROPAGATE) &&
				(m.param.pword[6] == VLP_REMOTE_INSTANCE_OK)) {
				interp->RInstances[on] = m.param.pword[7];
				break;
			}
			read(net_sock, &m, sizeof(MESSAGE));
		}

		Net_Notify->setEnabled(TRUE);

		/*bzero(&m, sizeof(MESSAGE));*/
		m.msg_type = MSG_VLP;
		m.param.pword[0] = VLP_REMOTE_INSTANCE_HERE;
		m.param.pword[1] = interp->RInstances[on];
		write(interp->sock, &m, sizeof(MESSAGE));
	} else { /* There is no such a node! */
		char s[255];
		sprintf(s, "Warning: Node number %d not found!", on); 
		WriteMessage(s);
		WriteMessage("Allocating O-process on the local node");
		bzero(&m, sizeof(MESSAGE));
		m.msg_type = MSG_VLP;
		m.param.pword[0] = VLP_REMOTE_INSTANCE_HERE;
		m.param.pword[1] = interp->ID;
		write(interp->sock, &m, sizeof(MESSAGE));
	}
}

/**
 * Closes all remote instances
 */
void QKernel::CloseInstances(InterpEntry *e)
{
	MESSAGE msg;
	int i;

	if (info_messages)
		WriteMessage("Closing remote instances");

	for(i=0; i < MAXINSTANCES; i++)
		if (e->RInstances[i]>=0) {
			msg.msg_type = MSG_NET;
			msg.param.pword[0] = NET_PROPAGATE;
			msg.param.pword[1] = MSG_VLP;
			msg.param.pword[2] = NodeNumber;
			msg.param.pword[4] = i;
			msg.param.pword[6] = VLP_CLOSE_INSTANCE;
			msg.param.pword[7] = e->RInstances[i];
			write(net_sock, &msg, sizeof(MESSAGE));
		}
}

/**
 * Displays information about virtual machine
 */
void QKernel::on_actionInfo_triggered()
{
	MESSAGE m;

	WriteMessage(CharLine);
	WriteMessage("### Virtual Machine Information ###");
	m.msg_type = MSG_NET;
	m.param.pword[0] = NET_GET_INFO;
	write(net_sock, &m, sizeof(MESSAGE));
	wait_for_info = TRUE;
}

}
}

/**
 * Program main function
 * All program arguments but the first one (argv[0]: program name) are saved and
 * passed to all dependent programs on their invocation.
 * @param argc Number of program arguments
 * @param argv Program arguments
 */
int main(int argc, char **argv)
{
	XInitThreads();

	QApplication * app = new QApplication(argc, argv);
	loglan::vlp::QKernel kernel(argc, argv);
	kernel.show();
	
	return app->exec();
}